# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .expression import Expression


class State:

    def __init__(self, name: str, node_name: str,
                 dynamic: Expression, initial: Expression):
        self.node_name = node_name
        self.name = name
        self.dynamic = dynamic
        self.initial = initial

    def get_node_name(self) -> str:

        return self.node_name

    def get_name(self) -> str:

        return self.name

    def get_dynamic(self) -> Expression:

        return self.dynamic

    def get_init(self) -> Expression:

        return self.initial


class Action:

    def __init__(self, name: str, node_name: str,
                 lower_bound: Expression, upper_bound: Expression):
        self.name = name
        self.node_name = node_name
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def get_node_name(self) -> str:

        return self.node_name

    def get_name(self) -> str:

        return self.name

    def get_lower_bound(self) -> Expression:

        return self.lower_bound

    def get_upper_bound(self) -> Expression:

        return self.upper_bound


class Sizing(Action):

    def __init__(self, name: str, node_name: str,
                 lower_bound: Expression, upper_bound: Expression):

        Action.__init__(self, name, node_name, lower_bound, upper_bound)


class Auxiliary:

    def __init__(self, name: str, node_name: str, definition: Expression):
        self.name = name
        self.node_name = node_name
        self.definition = definition

    def get_name(self) -> str:

        return self.name

    def get_definition(self) -> Expression:

        return self.definition


class MDPObjective:
    def __init__(self, node_name: str, expression: Expression):
        self.node_name = node_name
        self.expression = expression

    def get_node_name(self):
        return self.node_name

    def get_expression(self):
        return self.expression


class MDP:

    def __init__(self, states: list, actions: list,
                 sizing: list, auxiliaries: list, objectives: list):

        # states -> list of State objects
        # actions -> list of Action objects
        # sizing -> list of Sizing objects
        # auxiliaries -> list of Auxiliary objects

        self.sizing = sizing
        self.states = states
        self.actions = actions
        self.auxiliaries = auxiliaries
        self.objectives = objectives

    def get_objectives(self) -> list:
        return self.objectives

    def get_states_variables(self) -> list:

        return self.states

    def get_auxiliary_variables(self) -> list:

        return self.auxiliaries

    def get_sizing_variables(self) -> list:

        return self.sizing

    def get_actions_variables(self) -> list:

        return self.actions



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from gboml.compiler.utils import error_


class Hyperedge:

    def __init__(self, name, parameters=None, expressions=None,
                 constraints=None, line=0):

        self.name = name
        self.parameters = parameters
        self.constraints = constraints
        self.expressions = expressions
        self.parameter_dict = None
        self.nb_param = len(parameters)
        self.nb_constraint_matrix = 0
        self.nb_constraints = len(constraints)
        self.constr_factors = []
        self.c_triplet_list = []
        self.variables_used = {}
        self.line = line
        self.names_changes = []
        self.parameters_changes = []
        self.constraints_data = {}

    def get_line(self):
        return self.line

    def get_name(self):

        return self.name

    def rename(self, new_name):

        self.name = new_name

    def get_number_parameters(self):

        return self.nb_param

    def get_variables_used(self):

        return self.variables_used

    def get_expressions(self):
        return self.expressions

    def set_constraints_data(self, constr_data):
        self.constraints_data = constr_data

    def get_constraints_data(self):
        return self.constraints_data

    def add_constraint(self, constr):

        self.constraints.append(constr)

    def remove_constraint(self, constraint):

        self.constraints.remove(constraint)

    def set_constraints(self, constraints):

        self.constraints = constraints

    def get_constraints(self):

        return self.constraints

    def set_names_changes(self, changes):
        self.names_changes = changes

    def get_names_changes(self):
        return self.names_changes

    def add_name_change(self, change):
        self.names_changes.append(change)

    def set_parameters_changes(self, changes):
        self.parameters_changes = changes

    def add_parameter_change(self, change):
        self.parameters_changes.append(change)

    def get_parameters_changes(self):
        return self.parameters_changes

    def set_constraints_matrix(self, list_matrix):

        self.nb_constraint_matrix += len(list_matrix)
        self.c_triplet_list = list_matrix

    def get_constraints_matrix(self):

        return self.c_triplet_list

    def free_factors_constraints(self):
        self.constr_factors = None

    def get_number_expanded_constraints(self):

        return self.nb_constraint_matrix

    def get_number_constraints(self):

        return self.nb_constraints

    def add_parameters(self, param):

        self.parameters.append(param)

    def set_parameters(self, parameters):

        self.parameters = parameters

    def get_parameters(self):

        return self.parameters

    def set_parameter_dict(self, param):

        param = param.copy()
        if "global" in param:
            param.pop("global")
        if "GLOBAL" in param:
            param.pop("GLOBAL")
        if "T" in param:
            param.pop("T")
        self.parameter_dict = param

    def get_parameter_dict(self):

        return self.parameter_dict

    def get_dictionary_parameters(self):

        parameters = self.parameters
        all_parameters = dict()
        reserved_names = ["t", "T"]
        for param in parameters:

            name = param.get_name()
            if name in reserved_names:
                error_("Semantic error, variable named " + str(name) +
                       " is not allowed at line " + str(param.get_line()))
            if name not in all_parameters:

                all_parameters[name] = param
            else:

                error_("Semantic error, redefinition of variable "
                       + str(name) + " at line " + str(param.get_line()))

        return all_parameters

    def set_constraint_factors(self, fact_list):

        used_variables = {}
        for constr in fact_list:
            var_list = constr.variables
            nodes_in_constraint = set()
            for node_name, var_name in var_list:
                if node_name not in nodes_in_constraint:
                    nodes_in_constraint.add(node_name)

                if node_name not in used_variables:
                    list_var = [var_name]
                    used_variables[node_name] = list_var
                else:
                    list_var = used_variables[node_name]
                    if var_name not in list_var:
                        list_var.append(var_name)

            if len(nodes_in_constraint) == 1:
                print("Warning : Hyperedge constraint using "
                      "variables from a single node at line "
                      + str(constr.get_line()))

        self.variables_used = used_variables
        self.constr_factors = fact_list

    def get_constraint_factors(self):

        return self.constr_factors


class Attribute:
    """
    Attribute object is a structure composed of 
    - a node object
    - The node's name 
    - a variable name 
    """

    def __init__(self, name_node: str, name_variable=None, line: int = 0):

        assert type(name_node) == str, \
            "Internal error: Attribute node name of unknown type"
        self.node = name_node
        self.attribute = name_variable
        self.node_object = None  # POINTER to corresponding node object
        self.line = line

    def get_attribute(self):

        return self.attribute

    def get_node_field(self):

        return self.node

    def get_line(self):

        return self.line

    def __str__(self):
        
        string = ""
        if self.node_object is not None:

            string += '['
        string += str(self.node)
        if self.attribute is not None:

            string += '.'+str(self.attribute)
        if self.node_object is not None:

            string += ','+str(self.node_object.name)+']'
        
        return string

    def compare(self, attr):
        
        if self.node == attr.node and self.attribute == attr.attribute:

            return True
        
        return False

    def set_node_object(self, n_object):

        self.node_object = n_object

    def get_node_object(self):
        
        return self.node_object



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .version import __version__



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML compiler main file, compiles input file given in command line.

GBOML is an algebraic modelling language developed at the UNIVERSITY OF LIEGE.
This compiler takes GBOML input files and converts them into matrices to send
to solvers. Furthermore, once the problem solved, it outputs the results in an
understandable formalism similar to the input file.

  Typical usage example:

   $ python main.py gboml_file.txt --solver --output_type
  where:
    gboml_file is the file we want to compile
    --solver can either be linprog, cplex, clp/cbc, gurobi, xpress
    --output_type can either be csv or json

Several other options exists and can be retrieved by writing :
  python main.py -h
"""


from gboml import main


if __name__ == '__main__':
    main()



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Symbol
from gboml.compiler.utils import error_, list_to_string
from .link import Attribute
from .identifier import Identifier
from .error import WrongUsage
import copy


class Expression(Symbol):
    """
    Expression object is a tree like structure : 
    Its internal nodes are made up of 
    - an operator as type 
    - Children nodes of type Expression
    Its leafs are made up of 
    - a fixed type -> "literal"
    - no children nodes
    - a name field -> containing an evaluable unit 
    (number, identifier, etc)
    """

    def __init__(self, node_type: str, name=None, line: int = 0):

        assert type(node_type) == str, \
            "Internal error: expected string for Expression type"
        assert node_type in \
               ["+", "-", "/", "*", "**", "u-", "mod", "literal", "sum"],\
               "Internal error: unknown type for Expression"
        Symbol.__init__(self, name, node_type, line)
        self.parent = None
        self.children: list = []
        self.leafs = None
        self.condition = None
        self.time_interval = None

    def __str__(self) -> str:
        
        if self.type != "literal":

            string = '['+str(self.type)
            if self.name != "":

                string += " , "+str(self.name)
            if len(self.children) == 0:

                string += ']'
            else:

                string += "[" + list_to_string(self.children) + "]]"
        else:

            string = str(self.name)

        return string

    def free(self):

        for child in self.children:

            child.free()
        self.children = []
        if self.time_interval is not None:

            self.time_interval = None
        self.time_interval = None
        self.condition = None
        self.name = None
        self.parent = None
        self.type = None
        self.line = None

    def get_children(self):

        return self.children

    def set_children(self, children):

        self.children = children

    def get_nb_children(self):
        
        return len(self.children)

    def get_parent(self):

        return self.parent

    def add_parent(self, p):

        self.parent = p

    def add_child(self, child):

        child.add_parent(self)
        self.children.append(child)
    
    def set_leafs(self, leaves):

        self.leafs = leaves

    def get_leafs(self):
        
        if self.leafs is None:

            self.leafs = self.find_leafs()

        return self.leafs
    
    def set_time_interval(self, time_interv):

        self.time_interval = time_interv

    def get_time_interval(self):

        return self.time_interval

    def set_condition(self, cond):

        self.condition = cond

    def get_condition(self):

        return self.condition

    def check_time(self, definitions):

        predicate = True
        if self.condition is not None:
            predicate = self.condition.check(definitions)

        return predicate

    def expanded_leafs(self, definitions):

        expr_type = self.get_type()
        all_leafs = []
        if expr_type == "literal":

            return [self]
        elif expr_type == "sum":

            name_index = self.time_interval.get_index_name()
            range_index = self.time_interval.get_range(definitions)
            expr_list = self.get_children()
            original_expr = expr_list[0]
            if name_index in definitions or name_index == "t":

                error_("Already defined index name : %s" % name_index)

            for k in range_index:

                expr_copy = copy.copy(original_expr)
                new_expr = expr_copy.replace(name_index, k, definitions)
                new_leafs = new_expr.expanded_leafs(definitions)
                for leaf in new_leafs:

                    leaf.add_replacement(name_index, k)
                all_leafs += new_leafs

        else:

            children = self.get_children()
            for child in children:

                leafs = child.expanded_leafs(definitions)
                all_leafs += leafs

        return all_leafs

    def find_leafs(self):

        all_children = []
        if self.type == "literal" or self.type == "sum":

            all_children = [self]
        else:

            children = self.get_children()
            for child in children:

                all_children += child.get_leafs()
        
        return all_children

    def __copy__(self):

        copy_name = copy.copy(self.get_name())
        copy_type = copy.copy(self.get_type())

        expr_copy = Expression(copy_type, copy_name, self.get_line())
        children = self.get_children()
        new_children = [copy.copy(child) for child in children]
        time_interval = copy.copy(self.time_interval)
        expr_copy.set_children(new_children)
        expr_copy.set_time_interval(time_interval)
        
        return expr_copy

    def replace_basic_parameters(self, definitions):

        expr_type = self.get_type()
        name = self.get_name()

        if expr_type == "literal":

            if type(name) == Identifier:

                identifier = name
                id_node_name = identifier.get_node_name()
                id_name = identifier.get_name()
                id_type = identifier.get_type()
                parameter_type = ""
                to_replace = False
                value = 0

                if id_type != "basic":
                    WrongUsage("function replace basic parameters can only be "
                               "used for basic parameters")

                if id_node_name == "" and id_name in definitions:
                    parameter = definitions[id_name]
                    parameter_type = parameter.get_type()
                    value = parameter.get_value()
                    to_replace = True

                elif id_node_name in definitions and \
                        id_name in definitions[id_node_name]:
                    parameter = definitions[id_node_name][id_name]
                    parameter_type = parameter.get_type()
                    value = parameter.get_value()
                    to_replace = True

                if to_replace:
                    if parameter_type != 'expression':
                        WrongUsage("function replace basic parameters can "
                                   "only be used for basic parameters")

                    value = value[0]
                    self.name = value

        else:

            children = self.get_children()
            for child in children:
                child.replace_basic_parameters(definitions)

    def replace(self, name_index, value, definitions):

        expr_type = self.get_type()
        name = self.get_name()
        expr = self
        if expr_type == "literal":

            if type(name) == Identifier:

                identifier = name
                id_name = identifier.get_name()
                id_type = identifier.get_type()
                if id_type == "basic" and id_name == name_index:

                    expr = Expression('literal', value, line=self.line)
                elif id_type == "assign":

                    index_expr = identifier.get_expression()
                    index_expr = index_expr.replace(name_index, value,
                                                    definitions)
                    identifier.set_expression(index_expr)
                    self.set_name(identifier)
            elif type(name) == Attribute:

                attr = name
                identifier = attr.get_attribute()
                if identifier == 'assign':

                    index_expr = identifier.get_expression()
                    index_expr = index_expr.replace(name_index, value)
                    identifier.set_expression(index_expr)
                    self.set_name(identifier)
        else:

            children = self.get_children()
            new_children = []
            for child in children:

                new_child = child.replace(name_index, value, definitions)
                new_children.append(new_child)
            self.set_children(new_children)

        return expr

    def evaluate_expression(self, definitions: dict):

        # Get type, children and nb_children
        e_type = self.type
        children = self.get_children()
        nb_child = len(children)
        value = 0

        # if type is literal (leaf without child)
        if e_type == 'literal':

            if nb_child != 0:

                error_("INTERNAL ERROR : literal must have zero child, got "
                       + str(nb_child)+" check internal parser")

            # get identifier can either be a INT FLOAT ID or ID[expr]
            identifier = self.get_name()
            
            # retrieve value directly if FLOAT or INT
            if type(identifier) == float or type(identifier) == int:

                value = identifier
            elif type(identifier) == Attribute:

                key = identifier.get_node_field()
                if key not in definitions:

                    error_("Unknown Identifier "+str(identifier)+"at line "
                           + str(self.get_line()))
                inner_dict = definitions[key]
                inner_identifier = identifier.get_attribute()
                id_type = inner_identifier.get_type()
                id_name = inner_identifier.get_name()
                id_expr = inner_identifier.get_expression()
                if id_name not in inner_dict:

                    error_("Unknown Identifier "+str(identifier)+" at line "
                           + str(self.get_line()))
                if id_type == "basic":

                    value_vect = inner_dict[id_name]
                    if len(value_vect) != 1:

                        error_("INTERNAL error basic type should have "
                               "one value : "+str(identifier)
                               + " at line "+str(self.get_line()))
                    value = value_vect[0]
                elif id_type == "assign":

                    value_vect = inner_dict[id_name]
                    index = id_expr.evaluate_expression(definitions)

                    if len(value_vect) <= index:

                        error_("Wrong indexing in Identifier '"
                               + str(identifier) + "' at line, "
                               + str(self.get_line()))
                    value = value_vect[index]

            # else it is either ID or ID[expr]
            else:

                # get id type and set found to false
                identifier_type = identifier.get_type()
                identifier_name = identifier.get_name()
                identifier_expr = identifier.get_expression()
                identifier_node_name = identifier.get_node_name()
                dictionary_definitions = definitions

                if identifier_node_name in definitions:
                    dictionary_definitions = definitions[identifier_node_name]

                if identifier_name not in dictionary_definitions:
                    error_('Identifier "' + str(identifier)
                           + '" used but not previously defined, at line '
                           + str(self.get_line()))

                parameter = dictionary_definitions[identifier_name]
                vector_value = parameter.get_value()
                nb_values = len(vector_value)

                if identifier_type == "basic" and nb_values == 1:

                    value = vector_value[0]
                elif identifier_type == "assign":

                    index = identifier_expr.evaluate_expression(definitions)
                    if type(index) == float:

                        if index.is_integer() is False:

                            error_("Error: an index is a float: "
                                   + str(identifier)
                                   + 'at line '+str(identifier.get_line()))
                        index = int(round(index))
                    if index >= nb_values or index < 0:

                        error_("Wrong indexing in Identifier '"
                               + str(identifier)
                               + "' at line, "+str(self.get_line()))
                    value = vector_value[index]
                else:

                    error_("Wrong time indexing in Identifier '"
                           + str(identifier) + "' at line, "
                           + str(self.get_line()))
        # if expression type is unary minus
        elif e_type == 'u-':

            if nb_child != 1:
                error_("INTERNAL ERROR : unary minus must have one child, got "
                       + str(nb_child) + " check internal parser")

            # evaluate the children
            term1 = children[0].evaluate_expression(definitions)
            value = -term1

        elif e_type == "sum":
            from .parameter import Parameter
            time_int = self.time_interval.get_range(definitions)
            time_var = self.time_interval.get_index_name()
            if time_var in definitions:

                error_("ERROR: index "+str(time_var) +
                       " for loop already defined. Redefinition at line "
                       + str(self.line))
            sum_terms = 0
            condition_verified = False
            for i in time_int:

                index_parameter = Parameter(time_var, Expression("literal", i))
                index_parameter.set_value([i])
                definitions[time_var] = index_parameter
                if self.check_time(definitions):

                    condition_verified = True
                    term1 = children[0].evaluate_expression(definitions)
                    sum_terms += term1

            if not condition_verified:
                error_("ERROR: at expression :"+str(self)
                       + " there exists no value in sum range "
                         "that verifies that condition at line "+str(self.line))

            definitions.pop(time_var)
            value = sum_terms

        # MORE THAN one child
        else:

            if nb_child != 2:

                error_("INTERNAL ERROR : binary operators must have"
                       " two children, got "
                       + str(nb_child)+" check internal parser")
            term1 = children[0].evaluate_expression(definitions)
            term2 = children[1].evaluate_expression(definitions)
            if e_type == '+':

                value = term1 + term2
            elif e_type == '*':

                value = term1 * term2
            elif e_type == '/':

                value = term1/term2
            elif e_type == '-':

                value = term1-term2
            elif e_type == '**':

                value = term1**term2
            elif e_type == "mod":

                value = term1 % term2
            
            else:
                error_("INTERNAL ERROR : unexpected e_type "
                       + str(e_type)+" check internal parser")

        return value

    def evaluate_python_string(self, definitions: dict):
        # Discard precedence information before returning
        prec, value = self.evaluate_python_string_impl(definitions)
        return value

    def evaluate_python_string_impl(self, definitions: dict):
        # Return value is a pair of an int indicating precedence level
        # and a string representing the expression
        # Precedence levels are used to decide when a subexpression
        # must be put in parentheses
        # Parentheses are always required when a subexpression has
        # strictly lower precedence than the current operation
        # Precedence levels:
        # 0 - addition, binary minus
        # 1 - multiplications, division, modulo
        # 2 - unary minus
        # 3 - exponentiation
        # 4 - terminals

        # Get type, children and nb_children
        e_type = self.get_type()
        nb_child = self.get_nb_children()
        children = self.get_children()
        prec = None
        value = None

        # if expression type is unary minus
        if e_type == 'u-':
            if nb_child != 1:
                error_("INTERNAL ERROR : unary minus must have one child, got "
                       + str(nb_child)+" check internal parser")

            # evaluate the children
            prec1, term1 = children[0].evaluate_python_string_impl(definitions)
            if prec1 < 2:
                value = '-(' + term1 + ')'
            else:
                value = '-' + term1
            prec = 2

        # if type is literal (leaf without child)
        elif e_type == 'literal':
            if nb_child != 0:
                error_("INTERNAL ERROR : literal must have zero child, got "
                       + str(nb_child)+" check internal parser")

            # get identifier can either be a INT FLOAT ID or ID[expr]
            identifier = self.get_name()

            # retreive value directly if FLOAT or INT
            if type(identifier) == float or type(identifier) == int:
                value = str(identifier)
                prec = 4

            elif type(identifier) == Attribute:
                key = identifier.get_node_field()
                if key not in definitions:
                    error_("Unknown Identifier "+str(identifier)+" at line "
                           + str(self.get_line()))

                inner_dict = definitions[key]
                inner_identifier = identifier.get_attribute()

                id_type = inner_identifier.get_type()
                id_name = inner_identifier.get_name()
                id_expr = inner_identifier.get_expression()

                if id_name not in inner_dict:
                    error_("Unknown Identifier "+str(identifier)
                           + " at line "+str(self.get_line()))

                if id_type == "basic":
                    value_vect = inner_dict[id_name]
                    if len(value_vect) != 1:
                        error_("INTERNAL error basic type should "
                               "have one value : "+str(identifier)
                               + " at line "+str(self.get_line()))
                    value = str(value_vect[0])
                    prec = 4

                elif id_type == "assign":
                    value_vect = inner_dict[id_name]
                    index = id_expr.evaluate_expression(definitions)
                    if len(value_vect) <= index:
                        error_("Wrong indexing in Identifier '"
                               + str(identifier) + "' at line, "
                               + str(self.get_line()))
                    value = str(value_vect[index])
                    prec = 4

            # else it is either ID or ID[expr]
            else:
                # get id type and set found to false
                id_type = identifier.get_type()
                id_name = identifier.get_name()
                id_expr = identifier.get_expression()

                if not(id_name in definitions):
                    error_('Identifier "' + str(identifier)
                           + '" used but not previously defined, at line '
                           + str(self.get_line()))

                vector_value = definitions[id_name]
                nb_values = len(vector_value)

                if id_type == "basic" and nb_values == 1:
                    value = str(vector_value[0])
                    prec = 4

                elif id_type == "assign":
                    index = id_expr.evaluate_expression(definitions)

                    if type(index) == float:
                        if not index.is_integer():
                            error_("Error: an index is a float: '"
                                   + str(identifier) + "' at line "
                                   + str(identifier.get_line()))
                        index = int(round(index))

                    if index >= nb_values or index < 0:
                        error_("Wrong indexing in Identifier '"
                               + str(identifier) + "' at line, "
                               + str(self.get_line()))

                    value = str(vector_value[index])
                    prec = 4

                else:
                    error_("Wrong time indexing in Identifier '"
                           + str(identifier) + "' at line, "
                           + str(self.get_line()))

        elif e_type == 'sum':
            error_("DEBUG ERROR: Python string conversion for sum "
                   "not implemented")

        # MORE THAN one child
        else:

            if nb_child != 2:
                error_("INTERNAL ERROR : binary operators must have "
                       "two children, got "+str(nb_child)
                       + " check internal parser")

            prec1, term1 = children[0].evaluate_python_string_impl(definitions)
            prec2, term2 = children[1].evaluate_python_string_impl(definitions)
            if e_type == '+':
                if prec1 < 0:
                    value = '(' + term1 + ')'
                else:
                    value = term1
                value = value + '+'
                if prec2 < 0:
                    value = value + '(' + term2 + ')'
                else:
                    value = value + term2
                prec = 0
            elif e_type == '-':
                if prec1 < 0:
                    value = '(' + term1 + ')'
                else:
                    value = term1
                value = value + '-'
                if prec2 < 0:
                    value = value + '(' + term2 + ')'
                else:
                    value = value + term2
                prec = 0
            elif e_type == '*':
                if prec1 < 1:
                    value = '(' + term1 + ')'
                else:
                    value = term1
                value = value + '*'
                if prec2 < 1:
                    value = value + '(' + term2 + ')'
                else:
                    value = value + term2
                prec = 1
            elif e_type == '/':
                if prec1 < 1:
                    value = '(' + term1 + ')'
                else:
                    value = term1
                value = value + '/'
                if prec2 < 1:
                    value = value + '(' + term2 + ')'
                else:
                    value = value + term2
                prec = 1
            elif e_type == "mod":
                if prec1 < 1:
                    value = '(' + term1 + ')'
                else:
                    value = term1
                value = value + '%'
                if prec2 < 1:
                    value = value + '(' + term2 + ')'
                else:
                    value = value + term2
                prec = 1
            elif e_type == '**':
                if prec1 < 3:
                    value = '(' + term1 + ')'
                else:
                    value = term1
                value = value + '**'
                if prec2 < 3:
                    value = value + '(' + term2 + ')'
                else:
                    value = value + term2
                prec = 3
            else:
                error_("INTERNAL ERROR : unexpected e_type "
                       + str(e_type)+" check internal parser")

        return prec, value



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .parent import Type
from .expression import Expression


class Condition(Type):
    """
    Condition object is a tree like structure where:
    - each node is defined by an operator 
    - internal nodes have children of Condition objects
    - leaf nodes have children of Expressions type
    """

    def __init__(self, type_id: str, children: list, line: int = 0) -> None:
    
        assert type(type_id) == str, \
            "Internal error: expected string for condition type"
        assert type_id in \
               ["and", "not", "or", "==", "!=", "<", ">", "<=", ">="], \
               "Internal error: unknown type for condition"
        Type.__init__(self, type_id, line)
        self.type = type_id
        self.children = children

    def __str__(self) -> str:
    
        string = "["+str(self.type)
        for child in self.children:
            string += ','+str(child)
        string += ']'
    
        return string

    def get_children(self) -> list:
    
        return self.children
    
    def check(self, definitions: dict) -> bool:
    
        predicate = False
        if type(self.children[0]) == Condition:

            predicate_0 = self.children[0].check(definitions)
            predicate_1 = False
            if len(self.children) == 2:

                predicate_1 = self.children[1].check(definitions)
            if self.type == 'and':

                predicate = (predicate_0 and predicate_1)
            elif self.type == 'or':

                predicate = (predicate_0 or predicate_1)
            elif self.type == "not":

                predicate = (not predicate_0)
                
        elif type(self.children[0]) == Expression:
            value0 = self.children[0].evaluate_expression(definitions)
            value1 = self.children[1].evaluate_expression(definitions)

            if self.type == "==":
                predicate = (value0 == value1)
            elif self.type == "<=":
                predicate = (value0 <= value1)
            elif self.type == ">=":
                predicate = (value0 >= value1)
            elif self.type == ">":
                predicate = (value0 > value1)
            elif self.type == "<":
                predicate = (value0 < value1)
            elif self.type == "!=":
                predicate = (value0 != value1)

        return predicate



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML version.

The graph-based optimization modeling language (GBOML) version of the language.
"""

__version__ = "0.1.0"



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""CLP Solver file, contains the interface to CLP/CBC solver .

Takes the matrix A, the vector b and vector c as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the CLP/CBC solver.

  Typical usage example:

   solution, objective, status, solver_info = clp_solver(matrix_a,
                                                        vector_b,
                                                        vector_c,
                                                         objective_offset,
                                                         name_tuples)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""

import numpy as np
from scipy.sparse import coo_matrix
from gboml.compiler.utils import flat_nested_list_to_two_level


def clp_solver(matrix_a: coo_matrix, vector_b: np.ndarray, vector_c: np.ndarray,
               objective_offset: float,
               name_tuples: list) -> tuple:
    """clp_solver

        takes as input the matrix A, the vectors b and c. It returns the
        solution of the problem : min c^T * x s.t. A * x <= b found by
        the clp/cbc solver

        Args:
            A -> coo_matrix of constraints
            b -> np.ndarray of independent terms of each constraint
            c -> np.ndarray of objective vector
            objective_offset -> float of the objective offset
            name_tuples -> dictionary of <node_name variables> used to
                           get the type

        Returns:
            solution -> np.ndarray of the flat solution
            objective -> float of the objective value
            status -> solution status
            solver_info -> dictionary of solver information

    """

    try:

        from cylp.cy import CyCbcModel, CyClpSimplex
        from cylp.py.modeling.CyLPModel import CyLPModel, CyLPArray
    except ImportError:

        print("Warning: Did not find the CyLP package")
        exit(0)

    nvars = np.shape(vector_c)[1]
    additional_var_bool = False

    # CyLPModel with one variable creates an error
    # To go around this problem, we create another variable
    # which is never used and add it to the model
    if nvars == 1:

        additional_var_bool = True
        nvars = 2
        line, col = matrix_a.shape
        matrix_a = coo_matrix((matrix_a.data, (matrix_a.row, matrix_a.col)),
                              shape=(line, col+1))
        vector_c = np.append(vector_c[0], [0])

    # Build CLP model
    model = CyLPModel()
    c = CyLPArray(vector_c)
    variables = model.addVariable('variables', nvars, isInt=False)
    model.addConstraint(matrix_a * variables <= vector_b)
    model.objective = c * variables
    s = CyClpSimplex(model)
    flat_name_tuples = flat_nested_list_to_two_level(name_tuples)

    for index, _, var_type, var_size in flat_name_tuples:

        if var_type == "integer":

            s.setInteger(variables[index:index+var_size])
        if var_type == "binary":

            s.setInteger(variables[index:index+var_size])
            s += 0.0 <= variables[index:index+var_size] <= 1

    # Get cbc equivalent of that build clp model to exploit variable types
    cbc_model = s.getCbcModel()

    # Solve the model
    cbc_model.solve()

    # Gather and retrieve solver information
    solver_info = {"name": "clp", "algorithm": "primal simplex",
                   "status": cbc_model.status}
    solution = None
    objective = None

    status_code = cbc_model.status

    if status_code == "solution":

        status = "optimal"
        solution = cbc_model.primalVariableSolution['variables']

        # if we added an additional variable artificially
        # we remove it
        if additional_var_bool:

            solution = solution[0:len(solution)-1]
        objective = cbc_model.objectiveValue + objective_offset
    elif status_code == "unset":

        status = "unbounded"
        objective = float('-inf')
    elif status_code == 'relaxation infeasible':

        status = "infeasible"
        objective = float("inf")
    else:

        status = "unknown"

    return solution, objective, status, solver_info



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


import ply.yacc as yacc  # type: ignore
from copy import deepcopy
import os

from .gboml_lexer import tokens, lexer
from .utils import check_file_exists, error_, move_to_directory
from .classes import Time, Expression, Variable, Parameter, Program, Objective,\
    Node, Identifier, Constraint, Condition, TimeInterval, Hyperedge

list_opened_files = []
cache_graph = {}

# precedence rules from least to highest priority
# with associativity also specified

precedence = (  # Unary minus operator
    ('left', 'OR'),
    ('left', 'AND'),
    ('right', 'NOT'),
    ('nonassoc', 'EQUAL', 'LEQ', 'BEQ', 'BIGGER', 'LOWER', 'NEQ'),
    ('left', 'PLUS', 'MINUS'),
    ('left', 'MULT', 'DIVIDE'),
    ('right', 'UMINUS'),
    ('left', 'POW'),
    )


# Start symbol

def p_start(p):
    """start : time global program"""

    timehorizon = p[1]
    global_parameters = p[2]
    list_node, list_hyperlink = p[3]
    p[0] = Program(list_node, global_param=global_parameters,
                   timescale=timehorizon, links=list_hyperlink)


def p_global(p):
    """global : GLOBAL define_parameters
              | empty"""

    # if global parameters have been defined
    if len(p) == 3:

        global_parameters = p[2]
        p[0] = global_parameters

    # else no global parameters
    else:

        p[0] = []


def p_time(p):
    """time : TIME ID EQUAL expr SEMICOLON
            | empty"""

    # if a timehorizon has been defined
    if len(p) == 6:

        time_identifier = p[2]
        time_expression = p[4]
        p[0] = Time(time_identifier, time_expression, line=p.lineno(2))

    # if no timehorizon definition
    else:

        # force the timehorizon to 1 and emit warning
        expr = Expression('literal', 1, line=p.lineno(1))
        p[0] = Time("T", expr)
        print("WARNING: No timescale was defined ! Default : T = 1")


def p_program(p):
    """program : node program
                | hyperlink program
                | empty"""

    # if input file is empty
    if p[1] is None:

        p[0] = [[], []]

    # elif the first element is a node
    elif type(p[1]) == Node:

        list_node, list_hyperlink = p[2]
        list_node.append(p[1])
        p[0] = [list_node, list_hyperlink]

    # elif the first element is a hyperedge
    elif type(p[1]) == Hyperedge:

        list_node, list_hyperlink = p[2]
        list_hyperlink.append(p[1])
        p[0] = [list_node, list_hyperlink]


def p_hyperlink(p):
    """hyperlink : HYPEREDGE ID parameters expressions_definitions constraints
                 | HYPEREDGE ID EQUAL IMPORT list_of_id FROM FILENAME with_name_parameter_redefinition """

    if len(p) == 6:
        link_identifier = p[2]
        list_parameters = p[3]
        list_expressions = p[4]
        list_constraints = p[5]
        h_link = Hyperedge(link_identifier, list_parameters, list_expressions,
                           list_constraints, line=p.lexer.lineno)
        p[0] = h_link

    elif len(p) == 9:
        # RENAME Hyperedge
        link_identifier = p[2]
        imported_node_identifier = p[5]
        filename = p[7]
        list_parameters_redefinitions, list_name_redefinitions = p[8]
        old_dir, cut_filename = move_to_directory(filename)
        graph_filename = parse_file(cut_filename)
        returned_hyperedge = graph_filename.get(imported_node_identifier)
        returned_hyperedge = deepcopy(returned_hyperedge)
        if returned_hyperedge is None:
            error_("ERROR: In file " + str(filename)
                   + " there is no hyperedge named "
                   + str(imported_node_identifier))
        if type(returned_hyperedge) == Node:
            error_("ERROR: A node is imported as type Hyperedge at line "
                   + str(p.lexer.lineno))
        returned_hyperedge.set_names_changes(list_name_redefinitions)
        returned_hyperedge.set_parameters_changes(list_parameters_redefinitions)
        returned_hyperedge.rename(link_identifier)
        os.chdir(old_dir)
        p[0] = returned_hyperedge


def p_with_name_parameter_redefinition(p):
    """with_name_parameter_redefinition : WITH name_parameter_redefinition
                              | empty"""
    if len(p) == 2:
        p[0] = [[], []]
    elif len(p) == 3:
        p[0] = p[2]


def p_node(p):
    """node : NODE ID parameters program variables expressions_definitions constraints objectives
            | NODE ID EQUAL IMPORT list_of_id FROM FILENAME with_redefine_parameters_variables"""

    if type(p[3]) == list:
        sub_nodes, sub_edges = p[4]
        list_parameters = p[3]
        list_variables = p[5]
        list_expressions = p[6]
        list_constraints = p[7]
        list_objectives = p[8]
        node_identifier = p[2]
        p[0] = Node(node_identifier)
        p[0].set_line(p.lexer.lineno)
        p[0].set_sub_nodes(sub_nodes)
        p[0].set_sub_hyperedges(sub_edges)
        p[0].update_internal_dict()
        p[0].set_parameters(list_parameters)
        p[0].set_variables(list_variables)
        p[0].set_constraints(list_constraints)
        p[0].set_objectives(list_objectives)
        p[0].set_expressions(list_expressions)

    elif type(p[3]) == str:

        node_identifier = p[2]
        imported_node_identifier = p[5]
        filename = p[7]
        list_parameters, list_variables = p[8]
        old_dir, cut_filename = move_to_directory(filename)
        graph_filename = parse_file(cut_filename)
        returned_node = graph_filename.get(imported_node_identifier)
        returned_node = deepcopy(returned_node)
        if returned_node is None:
            error_("ERROR: In file "+str(filename)+" there is no node named "
                   + str(imported_node_identifier) +
                   " at line "+str(p.lexer.lineno))
        if type(returned_node) == Hyperedge:
            error_("ERROR: A hyperedge is imported as type node at line "
                   + str(p.lexer.lineno))
        returned_node.set_parameters_changes(list_parameters)
        returned_node.set_variables_changes(list_variables)
        returned_node.rename(node_identifier)
        os.chdir(old_dir)
        p[0] = returned_node


def p_list_of_id(p):
    """list_of_id : ID DOT list_of_id
                  | ID"""
    identifier_name = p[1]
    if len(p) == 2:
        p[0] = [identifier_name]
    elif len(p) == 4:
        list_id = p[3]
        list_id.insert(0, identifier_name)
        p[0] = list_id


def p_with_redefine_parameters_variables(p):
    """with_redefine_parameters_variables : WITH redefine_parameters_variables
                                          | SEMICOLON"""
    if len(p) == 2:
        p[0] = [[], []]
    else:
        p[0] = p[2]


def p_name_parameter_redefinition(p):
    """name_parameter_redefinition : ID ASSIGN ID SEMICOLON name_parameter_redefinition
                                   | ID ASSIGN ID SEMICOLON
                                   | parameter name_parameter_redefinition
                                   | parameter"""
    if len(p) >= 5:
        lhs_id = p[1]
        rhs_id = p[3]
        redefinition = [lhs_id, rhs_id, p.lineno(1)]
        list_parameters = []
        list_node_name_redefinition = []
        if len(p) == 6:
            list_parameters, list_node_name_redefinition = p[5]
        list_node_name_redefinition.append(redefinition)
        p[0] = [list_parameters, list_node_name_redefinition]

    else:
        parameter = p[1]
        list_parameters = []
        list_node_name_redefinition = []
        if len(p) == 3:
            list_parameters, list_node_name_redefinition = p[3]
        list_parameters.append(parameter)
        p[0] = [list_parameters, list_node_name_redefinition]


def p_redefine_parameters_variables(p):
    """redefine_parameters_variables : parameter redefine_parameters_variables
                                     | ID external_internal SEMICOLON redefine_parameters_variables
                                     | parameter
                                     | ID external_internal SEMICOLON"""

    if len(p) == 2:
        parameter = p[1]
        list_parameters = [parameter]
        list_variables = []
        p[0] = [list_parameters, list_variables]
    elif len(p) == 3:
        parameter = p[1]
        list_parameters, list_variables = p[2]
        list_parameters.append(parameter)
        p[0] = [list_parameters, list_variables]
    elif len(p) == 4:
        variable_name = p[1]
        variable_type = p[2]
        list_parameters = []
        list_variables = [[variable_name, variable_type, p.lineno(1)]]
        p[0] = [list_parameters, list_variables]
    elif len(p) == 5:
        variable_name = p[1]
        variable_type = p[2]
        list_parameters, list_variables = p[2]
        list_variables.append([variable_name, variable_type, p.lineno(1)])
        p[0] = [list_parameters, list_variables]


def p_parameters(p):
    """parameters : PARAM define_parameters
                  | empty"""

    # if no parameter has been defined
    if len(p) == 2:

        p[0] = []

    # else parameters have been defined
    else:

        list_parameters = p[2]
        p[0] = list_parameters


def p_define_parameters(p):
    """define_parameters : parameter define_parameters
                         | empty"""

    # if no parameter has been defined
    if len(p) == 2:

        p[0] = []

    # else parameters have been defined
    else:

        p[2].insert(0, p[1])
        p[0] = p[2]


def p_parameter(p):
    """parameter : ID EQUAL expr SEMICOLON
                 | ID EQUAL LCBRAC term more_values RCBRAC SEMICOLON
                 | ID EQUAL IMPORT FILENAME SEMICOLON"""

    parameter_id = p[1]

    # parameter defined by an expression
    if len(p) == 5:

        parameter_expression = p[3]
        p[0] = Parameter(parameter_id, parameter_expression, line=p.lineno(1))

    # parameter defined as a table
    elif len(p) == 8:

        list_values = p[5]
        last_value = p[4]
        p[0] = Parameter(parameter_id, None, line=p.lineno(1))
        list_values.insert(0, last_value)
        p[0].set_vector(list_values)

    # parameter defined via an import
    else:

        filename = p[4]
        p[0] = Parameter(parameter_id, filename, line=p.lineno(1))


def p_more_values(p):
    """more_values : COMMA term more_values
                    | empty"""

    # if no more values are used
    if len(p) == 2:

        p[0] = []

    # add additional value
    else:
        list_values = p[3]
        additional_value = p[2]
        list_values.insert(0, additional_value)
        p[0] = list_values


def p_variables(p):
    """variables : VAR define_variables"""

    p[0] = p[2]


def p_define_variables(p):
    """define_variables : external_internal type_var option_var COLON id SEMICOLON var_aux
                        | external_internal type_var option_var COLON id ASSIGN id SEMICOLON var_aux"""

    identifier_option = p[2]
    var_identifier = p[5]
    var_type = p[3]
    internal_external_keyword = p[1]

    # variable definition without assignment
    if len(p) == 8:
        list_variables = p[7]
        var_identifier.set_option(identifier_option)
        var = Variable(var_identifier, internal_external_keyword,
                       v_option=var_type, line=p.lineno(1))
        list_variables.insert(0, var)
        p[0] = list_variables

    # variable definition with assignment
    elif len(p) == 10:
        list_variables = p[9]
        child_variable = p[7]
        var_identifier.set_option(identifier_option)
        var = Variable(var_identifier, internal_external_keyword,
                       v_option=var_type, child_variable=child_variable,
                       line=p.lineno(1))
        list_variables.insert(0, var)
        p[0] = list_variables


def p_external_internal(p):
    """external_internal : INTERNAL
                         | EXTERNAL"""
    p[0] = p[1]


def p_type_var(p):
    """type_var : BINARY
                | CONTINUOUS
                | INTEGER
                | empty"""

    if p[1] is None:

        p[1] = "continuous"
    p[0] = p[1]


def p_option_var(p):
    """option_var : AUX
                  | ACTION
                  | SIZING
                  | STATE
                  | empty"""
    if p[1] is None:

        p[1] = "auxiliary"
    p[0] = p[1]


def p_var_aux(p):
    """var_aux :  define_variables
                | empty"""

    # case empty
    if p[1] is None:

        p[0] = []

    # case where more variables are defined
    else:

        p[0] = p[1]


def p_expressions_definition(p):
    """expressions_definitions : EXPRESSION expression_declaration expression_declaration_aux
                                | empty"""

    # empty case
    if len(p) == 2:
        p[0] = []

    # definition case
    else:
        list_expressions = p[3]
        additional_expression = p[2]
        list_expressions.insert(0, additional_expression)
        p[0] = list_expressions


def p_expression_declaration(p):
    """expression_declaration : ID EQUAL expr SEMICOLON"""
    p[0] = [p[1], p[3], p.lineno(1)]


def p_expression_declaration_aux(p):
    """expression_declaration_aux : expression_declaration expression_declaration_aux
                                  | empty"""
    if len(p) == 2:
        p[0] = []

    elif len(p) == 3:
        list_expressions = p[2]
        additional_expression = p[1]
        list_expressions.insert(0, additional_expression)
        p[0] = list_expressions


def p_constraints(p):
    """constraints : CONS constraints_aux
                   | empty"""
    # empty case
    if len(p) == 2:

        p[0] = []

    # definition case
    else:

        p[0] = p[2]


def p_constraints_aux(p):
    """constraints_aux : ID COLON define_constraints SEMICOLON constraints_aux
                       | ID COLON define_constraints SEMICOLON
                       | define_constraints SEMICOLON constraints_aux
                       | define_constraints SEMICOLON"""

    # last unnamed constraint definition
    if len(p) == 3:

        p[0] = []
        p[0].append(p[1])

    # add one unnamed constraint
    elif len(p) == 4:

        p[3].insert(0, p[1])
        p[0] = p[3]

    # last named constraint
    elif len(p) == 5:
        p[3].set_name(p[1])
        p[0] = []
        p[0].append(p[3])

    # add named constraint
    elif len(p) == 6:
        p[3].set_name(p[1])
        p[5].insert(0, p[3])
        p[0] = p[5]


def p_define_constraints(p):
    """define_constraints : expr DOUBLE_EQ expr time_loop condition
                          | expr LEQ expr time_loop condition
                          | expr BEQ expr time_loop condition"""

    # define a general constraint
    p[0] = Constraint(p[2], p[1], p[3], time_interval=p[4],
                      condition=p[5], line=p.lineno(2))


def p_condition(p):
    """condition : WHERE bool_condition
                 | empty"""

    # not empty case
    if len(p) == 3:

        p[0] = p[2]


def p_bool_condition(p):
    """bool_condition : bool_condition AND bool_condition
                      | NOT bool_condition
                      | bool_condition OR bool_condition
                      | LPAR bool_condition RPAR
                      | expr DOUBLE_EQ expr
                      | expr NEQ expr
                      | expr LOWER expr
                      | expr BIGGER expr
                      | expr LEQ expr 
                      | expr BEQ expr"""

    # AND OR PAR DOUBLE_EQ NEQ LOWER BIGGER LEQ BEQ cases
    if len(p) == 4:
        # AND OR DOUBLE_EQ NEQ LOWER BIGGER LEQ BEQ cases
        if type(p[2]) == str:

            children = [p[1], p[3]]
            p[0] = Condition(p[2], children, line=p.lineno(2))
        # PAR case
        else:

            p[0] = p[2]

    # NOT case
    else:

        p[0] = Condition(p[1], [p[2]], line=p.lineno(1))


def p_time_loop(p):
    """time_loop : FOR ID IN LBRAC expr COLON expr COLON expr RBRAC
                 | FOR ID IN LBRAC expr COLON expr RBRAC
                 | empty"""

    if len(p) == 11:

        p[0] = TimeInterval(p[2], p[5], p[9], p[7], p.lineno(2))
    elif len(p) == 9:
        step = Expression('literal', 1)
        p[0] = TimeInterval(p[2], p[5], p[7], step, p.lineno(2))


def p_objectives(p):

    """objectives : OBJ define_objectives
                  | empty"""

    if p[1] is None:

        p[0] = []
    else:

        p[0] = p[2]


def p_define_objectives(p):
    """define_objectives : MIN ID COLON expr time_loop condition SEMICOLON obj_aux
                         | MAX ID COLON expr time_loop condition SEMICOLON obj_aux
                         | MIN COLON expr time_loop condition SEMICOLON obj_aux
                         | MAX COLON expr time_loop condition SEMICOLON obj_aux"""

    if len(p) == 8:
        obj = Objective(p[1], p[3], p[4], p[5], line=p.lineno(1), name=None)
        p[7].insert(0, obj)
        p[0] = p[7]
    elif len(p) == 9:
        obj = Objective(p[1], p[4], p[5], p[6], line=p.lineno(1), name=p[2])
        p[8].insert(0, obj)
        p[0] = p[8]


def p_obj_aux(p):
    """obj_aux : define_objectives
               | empty"""

    if p[1] is None:

        p[0] = []
    else:

        p[0] = p[1]


def p_id(p):
    """id : ID
          | ID DOT ID
          | ID LBRAC expr RBRAC
          | ID DOT ID LBRAC expr RBRAC"""

    if len(p) == 2:
        # Rule ID
        p[0] = Identifier('basic', p[1], line=p.lineno(1))

    elif len(p) == 4:

        # Rule ID DOT ID
        p[0] = Identifier('basic', p[3], node_name=p[1], line=p.lineno(1))

    elif len(p) == 5:

        # Rule ID LBRAC expr RBRAC
        p[0] = Identifier('assign', p[1], expression=p[3], line=p.lineno(1))

    elif len(p) == 7:

        # Rule ID DOT ID LBRAC expr RBRAC
        p[0] = Identifier('assign', p[3], node_name=p[1], expression=p[5],
                          line=p.lineno(1))


def p_expr(p):
    """expr : expr PLUS expr %prec PLUS
            | expr MINUS expr %prec MINUS
            | expr MULT expr %prec MULT
            | expr DIVIDE expr %prec DIVIDE
            | MINUS expr %prec UMINUS
            | expr POW expr %prec POW
            | LPAR expr RPAR
            | MOD LPAR expr COMMA expr RPAR
            | SUM LPAR expr time_loop condition RPAR
            | term"""

    if len(p) == 4:
        
        # CASES + - * / ^ ()
        if type(p[2]) == str:

            p[0] = Expression(p[2], line=p.lineno(2))
            p[0].add_child(p[1])
            p[0].add_child(p[3])
        else:

            # ()
            p[0] = p[2]
    
    elif len(p) == 3:

        # CASE u-
        p[0] = Expression('u-', line=p.lineno(1))
        p[0].add_child(p[2])
    
    elif len(p) == 7:
        
        # CASE MODULO
        if p[1] == "mod":
            p[0] = Expression(p[1], line=p.lineno(1))
            p[0].add_child(p[3])
            p[0].add_child(p[5])

        if p[1] == "sum":
            # CASE SUM
            p[0] = Expression('sum', line=p.lineno(1))
            p[0].add_child(p[3])
            p[0].set_time_interval(p[4])
            p[0].set_condition(p[5])
    
    elif len(p) == 2:

        # CASE term
        p[0] = p[1]
    

def p_term(p):
    """term : INT
            | FLOAT
            | id"""

    p[0] = Expression('literal', p[1], line=p.lineno(1))
    if type(p[1]) == Identifier:
        p[0].set_line(p[1].get_line())


def p_empty(p):
    """empty :"""

    pass


# Error rule for syntax errors

def p_error(p):

    if p is not None:

        print('Syntax error: %d:%d: Unexpected token %s namely (%s)'
              % (p.lineno, find_column(p.lexer.lexdata, p),
                 p.type, str(p.value)))
    else:
        print('Syntax error: Expected a certain token got EOF(End Of File)')
    exit(-1)


def find_column(input_string, p):

    line_start = input_string.rfind('\n', 0, p.lexpos) + 1
    return p.lexpos - line_start + 1


def parse_file(name: str, cache=True) -> Program:
    """ parse_file

        takes as input a filename and converts it into a Program structure
        containing all the information

        Args:
            name (str) -> filename
            cache (bool) -> predicate whether to cache the imported nodes or not

        Returns:
            program -> program object containing all the information of the file

        Notes:
            When caching as only the filename is mapped to a program, conflicts
            can arise if a same filename is shared
            between two different files (on different paths typically)

    """
    # Build the parser
    global list_opened_files
    global cache_graph

    parser = yacc.yacc()
    check_file_exists(name)

    with open(name, 'r') as content:

        data = content.read()

    if name in list_opened_files:
        error_("ERROR: File "+str(name)
               + " has already been visited, there are loops in the import \n"
               + str(list_opened_files))

    if name in cache_graph and cache:
        return cache_graph[name]
    else:
        list_opened_files.append(name)
    # result = True

    result = parser.parse(data, lexer=lexer.clone())

    list_opened_files.pop(-1)

    cache_graph[name] = result

    return result



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from gboml.compiler.utils import list_to_string, error_, turn_to_dict
from .node import Node


def get_variable_name_size_offset_tuples_from_dict(variables_dict):
    tuple_names = []
    for node_name in variables_dict.keys():

        node_var_dict = variables_dict[node_name]
        all_tuples = []
        for var_name in node_var_dict.keys():
            if type(node_var_dict[var_name]) == dict:
                intermediate_dict = {var_name: node_var_dict[var_name]}
                name_index = \
                    get_variable_name_size_offset_tuples_from_dict(
                        intermediate_dict)

            else:
                variable = node_var_dict[var_name]
                identifier = variable.get_identifier()
                name_index = [identifier.get_index(),
                              var_name, identifier.get_option(),
                              identifier.get_size()]
            all_tuples.append(name_index)
        tuple_names.append([node_name, all_tuples])
    return tuple_names


class Program: 
    """
    Program object is composed of: 
    - a list of nodes
    - a Time object
    - a list of Links
    - a list of link constraint arrays
    """

    def __init__(self, vector_n, global_param=None, timescale=None, links=None):
        
        if global_param is None:

            global_param = []
        self.vector_nodes = vector_n
        self.time = timescale
        self.links = links
        self.global_param = global_param
        self.link_constraints = []
        self.nb_var_index = 0
        self.var_dict = {}
        self.link_list = []
        self.factor_links = []
        self.dict_nodes_links = turn_to_dict(self.vector_nodes+self.links)

    def __str__(self):
        
        string = "["+str(self.vector_nodes)
        if self.time is not None:

            string += ' , '+str(self.time)
        if self.links is None:

            string += ']'
        else:

            string += ' , '+str(self.links)+']'

        return string

    def to_string(self):
        
        string = "Full program\n"
        if self.time is not None:

            string += str(self.time)+"\n"
        string += 'All the defined nodes : \n'
        elements = self.vector_nodes
        for i in range(len(self.vector_nodes)):

            string += '\tName : ' + list_to_string(elements[i].get_name())+'\n'
            string += '\t\tParameters : ' \
                      + list_to_string(elements[i].get_parameters())+'\n'
            string += '\t\tVariables : ' \
                      + list_to_string(elements[i].get_variables())+'\n'
            string += '\t\tConstraints : ' \
                      + list_to_string(elements[i].get_constraints())+'\n'
            string += '\t\tObjectives : ' \
                      + list_to_string(elements[i].get_objectives())+'\n'
        string += '\nLinks predefined are : ' + str(self.links)
        
        return string

    def get(self, list_names):
        if type(list_names) == str:
            name = list_names
            if name not in self.dict_nodes_links:
                return -1
            else:
                return self.dict_nodes_links[name]
        elif (type(list_names) == list and len(list_names) >= 1) \
                or type(list_names) == tuple:
            actual_dict = self.dict_nodes_links
            actual_object = None
            for name in list_names:
                if name not in actual_dict:

                    return None
                else:
                    actual_object = actual_dict[name]
                    if type(actual_object) == Node:
                        actual_dict = actual_object.get_internal_dict()
                    else:
                        actual_dict = {}
            return actual_object

    def set_link_factors(self, factor_list):

        self.factor_links = factor_list

    def free_factors_objectives(self):
        for node in self.vector_nodes:
            node.free_factors_objectives()

    def free_factors_constraints(self):
        for node in self.vector_nodes:
            node.free_factors_constraints()
        self.factor_links = None

    def get_link_factors(self):

        return self.factor_links

    def get_global_parameters(self):

        return self.global_param

    def get_dict_global_parameters(self):

        dict_param = {}
        for param in self.global_param:

            name = param.get_name()
            if name in dict_param:

                error_("Global parameter "+str(name)+" already defined")
            else:

                dict_param[name] = param
        
        return dict_param

    def get_time(self):

        return self.time

    def get_nodes(self):
        
        return self.vector_nodes

    def get_number_nodes(self):

        return len(self.vector_nodes)

    def get_links(self):
        
        return self.links

    def set_time(self, timescale):
        
        self.time = timescale

    def set_vector(self, vector_n):
        
        self.vector_nodes = vector_n

    def set_link_constraints(self, c):
        
        self.link_constraints = c

    def get_link_constraints(self):
        
        return self.link_constraints
    
    def add_link_constraints(self, c):

        self.link_constraints.append(c)

    def get_number_constraints(self):
        
        sum_constraints = 0
        for node in self.vector_nodes:

            sum_constraints += node.get_nb_constraints_matrix()
        
        return sum_constraints 
    
    def check_objective_existence(self):
        
        nodes = self.vector_nodes
        found = False
        for node in nodes:

            objectives = node.get_objectives()
            nb_objectives = len(objectives)

            if nb_objectives != 0:

                found = True
                break
        if found is False:

            error_("ERROR: No objective function was defined")

    def get_nb_var_index(self):

        return self.nb_var_index

    def set_nb_var_index(self, index):

        self.nb_var_index = index

    def set_global_parameters(self, global_param):

        self.global_param = global_param

    def set_variables_dict(self, var_dict):

        self.var_dict = var_dict

    def get_variables_dict(self):

        return self.var_dict

    def get_tuple_name(self):

        return get_variable_name_size_offset_tuples_from_dict(self.var_dict)

    def add_var_link(self, tuple_list):
        link = []

        for element in tuple_list:

            [node_name, _, identifier, _] = element
            link.append([node_name, str(identifier)])
        self.link_list.append(link)

    def get_link_var(self):

        return self.link_list

    def get_first_level_constraints_decomposition(self):
        per_block_constraint_indexes = []
        start_index = 0
        current_index = 0
        nodes = self.vector_nodes
        hyperedges = self.links
        for node in nodes:
            current_index += node.get_number_expanded_constraints(True)
            per_block_constraint_indexes.append([start_index, current_index-1])
            start_index = current_index

        for hyperedge in hyperedges:
            current_index += hyperedge.get_number_expanded_constraints()
        per_block_constraint_indexes.append([start_index, current_index-1])
        return per_block_constraint_indexes



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .gboml_graph import GbomlGraph, VariableType
from .compiler import compile_gboml, compile_gboml_mdp
from .solver_api import scipy_solver, clp_solver,\
    cplex_solver, gurobi_solver, xpress_solver, dsp_solver
from .output import generate_json, generate_list_values_tuple, write_csv
from .gboml_main import main
from .version import __version__

__all__ = ["GbomlGraph", "VariableType", "main"]



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

class RedefinitionError(Exception):
    """Redefinition of protected element"""
    def __init__(self):
        self.message = "Attempt at overwriting a protected field"
        super().__init__(self.message)


class UndefinedIdentifier(Exception):
    """Undefined identifier used"""
    def __init__(self):
        self.message = "Identifier used but not defined"
        super().__init__(self.message)


class OutOfBoundIdentifier(Exception):
    """Trying to access an identifier """
    def __init__(self):
        self.message = "Out of bounds identifier assignment"
        super().__init__(self.message)


class WrongUsage(Exception):
    """Misusage of a function"""
    def __init__(self, message):
        super().__init__(message)



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""GBOML compiler main file, compiles input file given in command line.

GBOML is an algebraic modelling language developed at the UNIVERSITY OF LIEGE.
This compiler takes GBOML input files and converts them into matrices to send
to solvers. Furthermore, once the problem solved, it outputs the results in an
understandable formalism similar to the input file.

  Typical usage example:

   $ python main.py gboml_file.txt --solver --output_type
  where:
    gboml_file is the file we want to compile
    --solver can either be linprog, cplex, clp/cbc, gurobi, xpress
    --output_type can either be csv or json

Several other options exists and can be retrieved by writing :
  python main.py -h
"""

from .compiler import compile_gboml
from .solver_api import scipy_solver, clp_solver, \
    cplex_solver, gurobi_solver, xpress_solver, dsp_solver
from .output import generate_json, generate_list_values_tuple, write_csv

import argparse
import json
import numpy as np
import sys
from time import gmtime, strftime, time


def main():
    parser = argparse.ArgumentParser(allow_abbrev=False,
                                     description='Compiler and solver for the '
                                                 'generic system model language'
                                     )
    parser.add_argument("input_file", type=str)

    # Compiling info
    parser.add_argument("--lex", help="Prints all tokens found in input file",
                        action='store_const', const=True)
    parser.add_argument("--parse", help="Prints the AST", action='store_const',
                        const=True)
    parser.add_argument("--matrix", help="Prints matrix representation",
                        action='store_const', const=True)
    parser.add_argument("--nb_processes", help="Number of processes to use",
                        type=int)

    # Solver
    parser.add_argument("--clp", help="CLP solver", action='store_const',
                        const=True)
    parser.add_argument("--cplex", help="Cplex solver", action='store_const',
                        const=True)
    parser.add_argument("--linprog", help="Scipy linprog solver",
                        action='store_const', const=True)
    parser.add_argument("--gurobi", help="Gurobi solver",
                        action='store_const', const=True)
    parser.add_argument("--xpress", help="Xpress solver",
                        action='store_const', const=True)
    parser.add_argument("--dsp_de", help="DSP Extensive Form algorithm",
                        action='store_const', const=True)
    parser.add_argument("--dsp_dw", help="DSP Dantzig-Wolf algorithm",
                        action="store_const", const=True)

    # Output
    parser.add_argument("--csv", help="Convert results to CSV format",
                        action='store_const', const=True)
    parser.add_argument("--transposed_csv",
                        help="Convert results to CSV format sorted by columns",
                        action='store_const', const=True)
    parser.add_argument("--json", help="Convert results to JSON format",
                        action='store_const', const=True)
    parser.add_argument("--detailed_json",
                        help="Convert detailed results to JSON format",
                        action="store_const", const=True)
    parser.add_argument("--log", help="Get log in a file",
                        action="store_const", const=True)
    parser.add_argument("--output", help="Output filename", type=str)
    parser.add_argument("--opt", help="Optimization options filename", type=str)

    args = parser.parse_args()
    start_time = time()

    if args.input_file:
        if args.nb_processes is None:
            args.nb_processes = 1
        elif args.nb_processes <= 0:
            print("The number of processes must be strictly positive")
            exit()

        program, A, b, C, indep_terms_c, T, name_tuples = \
            compile_gboml(args.input_file, args.log, args.lex,
                          args.parse, args.nb_processes)

        print("All --- %s seconds ---" % (time() - start_time))
        C_sum = np.asarray(C.sum(axis=0), dtype=float)

        if args.matrix:
            print("Matrix A ", A)
            print("Vector b ", b)
            print("Vector C ", C_sum)

        objective_offset = float(indep_terms_c.sum())
        status = None

        constraints_additional_information = dict()
        variables_additional_information = dict()

        if args.linprog:

            x, objective, status, solver_info = \
                scipy_solver(A, b, C_sum, objective_offset, name_tuples)
        elif args.clp:

            x, objective, status, solver_info = \
                clp_solver(A, b, C_sum, objective_offset, name_tuples)
        elif args.cplex:

            x, objective, status, solver_info, \
             constraints_additional_information, \
             variables_additional_information = \
             cplex_solver(A, b, C_sum, objective_offset, name_tuples, args.opt)

        elif args.gurobi:

            x, objective, status, solver_info, \
             constraints_additional_information, \
             variables_additional_information = \
             gurobi_solver(A, b, C_sum, objective_offset, name_tuples, args.opt)

        elif args.xpress:

            x, objective, status, solver_info, \
             constraints_additional_information, \
             variables_additional_information = \
             xpress_solver(A, b, C_sum, objective_offset, name_tuples, args.opt)

        elif args.dsp_dw:
            x, objective, status, solver_info = \
                dsp_solver(A, b, C_sum, objective_offset, name_tuples,
                           program.get_first_level_constraints_decomposition(),
                           algorithm="dw")

        elif args.dsp_de:
            x, objective, status, solver_info = \
                dsp_solver(A, b, C_sum, objective_offset, name_tuples,
                           program.get_first_level_constraints_decomposition(),
                           algorithm="de")

        else:

            print("No solver was chosen")
            sys.exit()

        assert status in {"unbounded", "optimal", "feasible", "infeasible",
                          "error", "unknown"}

        if status == "unbounded":

            print("Problem is unbounded")
        elif status == "optimal":

            print("Optimal solution found")
        elif status == "feasible":

            print("Feasible solution found")
        elif status == "infeasible":

            print("Problem is infeasible")
        elif status == "error":

            print("An error occurred")
            exit()
        elif status == "unknown":

            print("Solver returned with unknown status")
        if args.output:

            filename = args.output
        else:

            filename_split = args.input_file.rsplit('.', 1)
            filename = filename_split[0]
            time_str = strftime("%Y_%m_%d_%H_%M_%S", gmtime())
            filename = filename + "_" + time_str

        if args.json or args.detailed_json:
            if args.json and args.detailed_json:
                print("Warning: Selected both json and "
                      "detailed json results in only detailed "
                      "json being saved")
            dictionary = dict()
            if args.json:
                dictionary = generate_json(program, solver_info, status, x,
                                           objective, C,
                                           indep_terms_c)
            if args.detailed_json:
                dictionary = generate_json(program, solver_info, status, x,
                                           objective, C,
                                           indep_terms_c,
                                           constraints_additional_information,
                                           variables_additional_information)
            try:
                with open(filename + ".json", 'w') as outfile:

                    json.dump(dictionary, outfile, indent=4)

                print("File saved: " + filename + ".json")
            except PermissionError:

                print("WARNING the file " + str(filename)
                      + ".json already exists and is open.")
                print("Was unable to save the file")
        if args.csv or args.transposed_csv:
            names_var_and_param, values_var_and_param = \
                generate_list_values_tuple(program, x)
            write_csv(filename + ".csv", names_var_and_param,
                      values_var_and_param, transpose=args.transposed_csv)
    else:

        print('ERROR : expected input file')
    print("--- %s seconds ---" % (time() - start_time))



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

"""Xpress Solver file, contains the interface to Xpress solver .

Takes the matrix A, the vector b and vector c as input of the problem
    min : c^T * X s.t. A * X <= b
and passes it to the xpress solver.

  Typical usage example:

   solution, objective, status, solver_info = xpress_solver(matrix_a, vector_b,
                                                            vector_c,
                                                            objective_offset,
                                                            name_tuples)
   print("the solution is "+str(solution))
   print("the objective found : "+str(objective))
"""

import numpy as np
from scipy.sparse import coo_matrix, csr_matrix
from gboml.compiler.utils import flat_nested_list_to_two_level


def xpress_solver(matrix_a: coo_matrix, vector_b: np.ndarray,
                  vector_c: np.ndarray, objective_offset: float,
                  name_tuples: list,
                  opt_file: str = None) -> tuple:
    """xpress_solver

        takes as input the matrix A, the vectors b and c. It returns
        the solution of the problem : min c^T * x s.t. A * x <= b found
        by the xpress solver

        Args:
            A -> coo_matrix of constraints
            b -> np.ndarray of independent terms of each constraint
            c -> np.ndarray of objective vector
            objective_offset -> float of the objective offset
            name_tuples -> dictionary of <node_name variables> used to get
                           the type
            opt_file -> optimization parameters file

        Returns:
            solution -> np.ndarray of the flat solution
            objective -> float of the objective value
            status -> solution status
            solver_info -> dictionary of solver information

    """

    try:

        import xpress as xp
    except ImportError:

        print("Warning: Did not find the CyLP package")
        exit(0)

    if opt_file is None:
        opt_file = 'gboml/solver_api/xpress.opt'

    # Generating the model
    model = xp.problem()
    matrix_a = matrix_a.astype(float)
    _, nb_columns = matrix_a.shape

    var_list = [xp.var(vartype=xp.continuous, lb=float('-inf'))
                for _ in range(nb_columns)]
    flat_name_tuples = flat_nested_list_to_two_level(name_tuples)
    is_milp = False
    for index, _, var_type, var_size in flat_name_tuples:
        if var_type == "integer":
            is_milp = True
            for i in range(var_size):
                var_list[index+i] = xp.var(vartype=xp.integer, lb=float('-inf'))
        elif var_type == "binary":
            is_milp = True
            for i in range(var_size):
                var_list[index+i] = xp.var(vartype=xp.binary)

    var_array = np.array(var_list)
    model.addVariable(var_array)
    nb_constraints, _ = matrix_a.shape
    csr_matrix_format_a = matrix_a.tocsr()
    csr_data, csr_indices, csr_ptr = csr_matrix_format_a.data, \
                                     csr_matrix_format_a.indices, \
                                     csr_matrix_format_a.indptr

    for i in range(nb_constraints):
        pt = slice(csr_ptr[i], csr_ptr[i+1])
        columns = csr_indices[pt]
        values = csr_data[pt]
        lhs_constraint = xp.Dot(np.array(values), var_array[columns])
        model.addConstraint(lhs_constraint <= vector_b[i])

    objective = xp.Dot(vector_c.reshape(-1), var_array) + objective_offset
    model.setObjective(objective)

    # Retrieve solver information
    option_info = {}
    try:

        with open(opt_file, 'r') as optfile:

            lines = optfile.readlines()
    except IOError:

        print("Options file not found")
    else:

        for line in lines:

            line = line.strip()
            option = line.split(" ", 1)
            if option[0] != "":
                try:
                    parinfo = getattr(model.controls, option[0])
                except AttributeError as e:
                    print("Skipping unknown option \'%s\'" % option[0])
                    continue
                if parinfo:

                    if len(option) == 2:

                        key = option[0]
                        try:
                            value = parinfo
                        except ValueError as e:

                            print("Skipping option \'%s\' with invalid "
                                  "given value \'%s\' (expected %s)"
                                  % (option[0], option[1], parinfo[1]))
                        else:

                            model.setControl(key, value)
                            option_info[key] = value
                    else:

                        print("Skipping option \'%s\' with no given value"
                              % option[0])
                else:

                    print("Skipping unknown option \'%s\'" % option[0])

    # Solve the model and generate output
    model.solve()
    solution = np.array(model.getSolution())
    objective = model.getObjVal()

    status = model.getProbStatus()
    if (status == 1 and not is_milp) or (is_milp and status == 6):
        status = "optimal"
    elif (status == 2 and not is_milp) or (is_milp and status == 5):
        status = "infeasible"
    else:
        status = "unknown"

    solver_info = {"name": "xpress", "algorithm": "unknown",
                   "status": status,
                   "options": option_info}

    constraints_additional_information = dict()
    variables_additional_information = dict()

    attributes_to_retrieve_constraints = [
        ["dual", model.getDual],
        ["slack", model.getSlack],
    ]
    attributes_to_retrieve_variables = [
        ["reduced_cost", model.getRCost],
    ]

    for name, function in attributes_to_retrieve_constraints:
        try:
            constraints_additional_information[name] = function()
        except RuntimeError:
            print("Unable to retrieve ", name, " information for constraints")

    for name, function in attributes_to_retrieve_variables:
        try:
            variables_additional_information[name] = function()
        except RuntimeError:
            print("Unable to retrieve ", name, " information for variables")

    return solution, objective, status, solver_info, \
           constraints_additional_information, \
           variables_additional_information



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).


"""Output file, generates the different types of output.

There are two possibilities of output, either json or csv. The two functions
generate_json and generate_pandas generate datastructures that can be written
into respectively a json and csv file.

  Typical usage example:

   json_output = generate_json(program, variable_names, solver_data,
                              status, solution, objective, c_matrix,
                              indep_terms_c,
                              objective_map)
   with open(filename+".json", 'w') as outfile:
        json.dump(dictionary, outfile, indent=4)

"""

import csv
from gboml.version import __version__


def convert_parameter_dict_to_values(parameter_name_object_dict: dict) -> dict:
    """convert_parameter_dict_to_values

        retrieves the values out of a dictionary of Parameter objects where
        the keys are the parameter names and returns a similar dictionary where
        the keys are the same parameter names with their value replaced by the
        corresponding parameter value

        Args:
            parameter_name_object_dict (dict): dictionary of <parameter name,
            Parameter objects>

        Returns:
            parameter_name_value_dict: dictionary of <parameter name,
            list of values>

    """

    parameter_name_value_dict = {}
    for parameter_name in parameter_name_object_dict.keys():
        parameter_name_value_dict[parameter_name] = \
            parameter_name_object_dict[parameter_name].get_value()
    return parameter_name_value_dict


def get_nodes_and_edges_model_information(nodes, edges):
    """get_nodes_and_edges_model_information

        recursively turning all the information relative to the model
        contained in a list of nodes and hyperedges

        Args:
            nodes (list <Node>) : list of nodes
            edges (list <Hyperedge>) : list of hyperedges

        Returns:
            nodes_information : node model information
            hyperedge_info : hyperedge model information

    """
    nodes_information = {}
    for node in nodes:
        node_data = dict()
        node_data["number_parameters"] = node.get_number_parameters()
        node_data["number_variables"] = node.get_number_variables()
        node_data["number_constraints"] = node.get_number_constraints()
        node_data["number_expanded_constraints"] = \
            node.get_number_expanded_constraints()
        node_data["number_objectives"] = node.get_number_objectives()
        node_data["number_expanded_objectives"] = \
            node.get_number_expanded_objectives()
        node_data["parameters"] = \
            convert_parameter_dict_to_values(node.get_parameter_dict())
        node_data["variables"] = node.get_variable_names()
        sub_nodes = node.get_sub_nodes()
        sub_edges = node.get_sub_hyperedges()
        sub_nodes_info, sub_hyperedge_info = \
            get_nodes_and_edges_model_information(sub_nodes, sub_edges)
        if sub_nodes_info != {}:
            node_data["sub_nodes"] = sub_nodes_info
        if sub_hyperedge_info != {}:
            node_data["sub_hyperedges"] = sub_hyperedge_info
        nodes_information[node.get_name()] = node_data

    hyperedge_info = {}
    for link in edges:
        link_data = dict()
        link_data["number_parameters"] = link.get_number_parameters()
        link_data["number_constraints"] = link.get_number_constraints()
        link_data["number_expanded_constraints"] = \
            link.get_number_expanded_constraints()
        link_data["parameters"] = \
            convert_parameter_dict_to_values(link.get_parameter_dict())
        link_data["variables_used"] = link.get_variables_used()
        hyperedge_info[link.get_name()] = link_data

    return nodes_information, hyperedge_info


def dict_values_in_nested_dict(nodes, hyperedges, solution,
                               c_matrix_time_solution,
                               constraint_info=None,
                               variables_info=None):
    """dict_values_in_nested_dict

        recursively turning all the information relative to the solution
        contained in a list of nodes and hyperedges

        Args:
            nodes (list <Node>) : list of nodes
            hyperedges (list <Hyperedge>) : list of hyperedges
            solution (numpy array) : array of the solution
            c_matrix_time_solution (numpy array) : array of the
                                                   objective_matrix * solution
                                                   + independent terms
            constraint_info (dict) : eventual additional information
                                     relative to the constraints
            variables_info (dict) : eventual additional information relative

        Returns:
            dict_solution (dict) : dictionary gathering and structuring
                                   all the information

    """
    if variables_info is None:
        variables_info = dict()
    if constraint_info is None:
        constraint_info = dict()
    dict_solution = {}

    for node in nodes:
        dict_node_solution = {}
        dict_variables_solution = {}
        dict_constraints_solution = {}
        node_name = node.get_name()
        variables_dict = node.get_dictionary_variables()

        for var_name, identifier in variables_dict.items():
            all_info_for_variable = {}
            start_index = identifier.get_index()
            size_variable = identifier.get_size()
            values_concerned = solution[start_index:
                                        (start_index + size_variable)]
            if type(values_concerned) != list:
                all_info_for_variable["values"] = values_concerned.flatten().tolist()
            else:
                all_info_for_variable["values"] = values_concerned

            for variable_info_name, variable_info_values \
                    in variables_info.items():
                if len(variable_info_values) > 1 and\
                        type(variable_info_values[0]) == list:
                    list_info = []
                    for value_info in variable_info_values:
                        list_info.append(value_info[start_index:
                                                    (start_index
                                                     + size_variable)])
                    all_info_for_variable[variable_info_name] = list_info
                else:
                    all_info_for_variable[variable_info_name] = \
                        variable_info_values[start_index:
                                             (start_index + size_variable)]
            dict_variables_solution[var_name] = all_info_for_variable

        dict_node_solution["variables"] = dict_variables_solution

        constraints_info_in_node = node.get_constraints_data()

        if constraint_info != {}:
            for constraint_name, indexes in constraints_info_in_node.items():
                all_info_for_constraint = {}
                for constraint_info_name, constraint_info_values in \
                        constraint_info.items():
                    all_info_for_constraint[constraint_info_name] = \
                        constraint_info_values[indexes]
                dict_constraints_solution[constraint_name] = \
                    all_info_for_constraint

            if dict_constraints_solution != {}:
                dict_node_solution['constraints'] = dict_constraints_solution

        named_objectives_info = {}
        unnamed_objectives_info = []
        named_and_unnamed_objectives_info = {}
        objectives_data = node.get_objectives_data()
        for i, objective_info in objectives_data.items():
            objective_type = objective_info["type"]
            indexes = objective_info["indexes"]
            objective_value = c_matrix_time_solution[indexes].sum()
            if objective_type == "max":
                objective_value = -objective_value

            if "name" in objective_info:
                name_objective = objective_info["name"]
                named_objectives_info[name_objective] = objective_value
            else:
                unnamed_objectives_info.append(objective_value)

        if named_objectives_info != {}:
            named_and_unnamed_objectives_info["named"] = named_objectives_info

        if unnamed_objectives_info:
            named_and_unnamed_objectives_info["unnamed"] = \
                unnamed_objectives_info

        if named_and_unnamed_objectives_info != {}:
            dict_node_solution["objectives"] = named_and_unnamed_objectives_info

        sub_nodes = node.get_sub_nodes()
        sub_hyperedges = node.get_sub_hyperedges()

        sub_elements = dict_values_in_nested_dict(sub_nodes,
                                                  sub_hyperedges,
                                                  solution,
                                                  c_matrix_time_solution,
                                                  constraint_info,
                                                  variables_info)
        if sub_elements != {}:
            dict_node_solution["sub_elements"] = sub_elements

        dict_solution[node_name] = dict_node_solution

    if constraint_info != {}:
        for hyperedge in hyperedges:
            hyperedge_name = hyperedge.get_name()
            dict_hyperedge_solution = {}
            dict_hyperedge_constraints_solution = {}
            constraints_info_in_node = hyperedge.get_constraints_data()
            for constraint_name, indexes in constraints_info_in_node.items():
                all_info_for_constraint = {}
                for constraint_info_name, constraint_info_values in \
                        constraint_info.items():
                    all_info_for_constraint[constraint_info_name] = \
                        constraint_info_values[indexes]
                dict_hyperedge_constraints_solution[constraint_name] = \
                    all_info_for_constraint

            if dict_hyperedge_constraints_solution != {}:
                dict_hyperedge_solution['constraints'] = \
                    dict_hyperedge_constraints_solution
                dict_solution[hyperedge_name] = dict_hyperedge_solution

    return dict_solution


def generate_json(program, solver_data, status, solution, objective, c_matrix,
                  indep_terms_c, constraint_info=None,
                  variables_info=None) -> dict:
    """generate_json

        Converts all the information contained in the inputs into one dictionary
        that can be dumped in a json file

        Args:
            program (Program): program object containing the augmented abstract
                               syntax tree
            solver_data (dict): dictionary containing the solver data
            status (str): status of the solver
            solution (array): flat array containing the problem's solution
            objective (float): value of the objective
            c_matrix (array): matrix of all the objectives
            indep_terms_c (array): array of all the independent terms of
                                   each objective
            constraint_info (dict): dictionary containing additional
                                    constraints information
            variables_info (dict): dictionary containing additional
                                   variables information

        Returns:
            gathered_data: dictionary containing all the gathered information

    """
    if variables_info is None:
        variables_info = dict()
    if constraint_info is None:
        constraint_info = dict()
    gathered_data = dict()
    gathered_data["version"] = __version__
    model_data = {}
    horizon = program.time.get_value()
    model_data["horizon"] = horizon
    model_data["number_nodes"] = program.get_number_nodes()
    model_data["global_parameters"] = \
        convert_parameter_dict_to_values(program.get_global_parameters())
    nodes, hyperlinks = \
        get_nodes_and_edges_model_information(program.get_nodes(),
                                              program.get_links())

    model_data["nodes"] = nodes
    model_data["hyperedges"] = hyperlinks

    gathered_data["model"] = model_data
    gathered_data["solver"] = solver_data

    solution_data = dict()
    solution_data["status"] = status
    solution_data["objective"] = objective

    if solution is not None:
        product = c_matrix * solution + indep_terms_c

        solution_data["elements"] = \
            dict_values_in_nested_dict(program.get_nodes(),
                                       program.get_links(),
                                       solution,
                                        product,
                                       constraint_info,
                                       variables_info)
    gathered_data["solution"] = solution_data

    return gathered_data


def flat_graph_and_add_node_prefix(nodes, hyperedges, solution, prefix=""):
    """flat_graph_and_add_node_prefix

        creates a list of the nodes and hyperedges variables & parameters
         and their value. The name of the variable and parameters is
         given as : prefix.parent.parent.

        Args:
            nodes (list<Node>): list of nodes
            hyperedges (list<Hyperedge>): list of hyperedges
            solution (array): flat array containing the problem's solution
            prefix (str): prefix to add to the variables and parameters names

        Returns:
            name_tuples: list of tuples <identifier, values>

    """
    name_tuples = []
    for node in nodes:
        variables_dict = node.get_dictionary_variables()
        node_name = prefix+node.get_name()
        for var_name, identifier in variables_dict.items():

            start_index = identifier.get_index()
            size_variable = identifier.get_size()
            identifier_name = node_name+"."+identifier.get_name()
            values = solution[start_index:
                              (start_index + size_variable)].flatten().tolist()
            name_tuples.append([identifier_name, values])

        parameters = node.get_parameter_dict()
        for parameter_name, parameter_object in parameters.items():

            values = parameter_object.get_value()
            parameter_full_name = str(node_name) + "." + str(parameter_name)
            name_tuples.append([parameter_full_name, values])

        sub_nodes = node.get_sub_nodes()
        sub_hyperedges = node.get_sub_hyperedges()
        name_tuples += flat_graph_and_add_node_prefix(sub_nodes,
                                                      sub_hyperedges,
                                                      solution,
                                                      node_name+".")

    for hyperedge in hyperedges:
        hyperedge_name = prefix+hyperedge.get_name()
        parameters = hyperedge.get_parameter_dict()
        for parameter_name, parameter_object in parameters.items():
            values = parameter_object.get_value()
            parameter_full_name = str(hyperedge_name) + "." \
                                  + str(parameter_name)
            name_tuples.append([parameter_full_name, values])

    return name_tuples


def generate_list_values_tuple(program, solution):
    """generate_pandas

        Converts all the information contained in the inputs into a tuple
        of names and values where values[i] is the list of values
        corresponding to names[i]

        Args:
            program (Program): program object containing the augmented
                               abstract syntax tree
            solution (array): flat array containing the problem's solution

        Returns:
            names: list of flatten names of all parameters and variables
            values : list of values

    """
    ordered_values = []
    names = []
    nodes = program.get_nodes()
    hyperedges = program.get_links()

    global_param: dict = program.get_global_parameters()
    for param in global_param.keys():
        values = global_param[param].get_value()
        full_name = "global." + str(param)
        names.append(full_name)
        ordered_values.append(values)

    name_value_tuples = flat_graph_and_add_node_prefix(nodes,
                                                       hyperedges,
                                                       solution,
                                                       prefix="")

    for name, values in name_value_tuples:
        names.append(name)
        ordered_values.append(values)

    return names, ordered_values


def write_csv(filename, names, values, transpose=False):

    if transpose is True:
        max_length = max([len(i) for i in values])
        column_ordered_values = []
        for i in range(max_length):
            row_i = []
            for value in values:
                if i < len(value):
                    row_i.append(value[i])
                else:
                    row_i.append(None)
            column_ordered_values.append(row_i)
        try:
            with open(filename, 'w', encoding='UTF8', newline="") as f:
                writer = csv.writer(f)
                writer.writerow(names)
                for row_i in column_ordered_values:
                    writer.writerow(row_i)
        except PermissionError:

            print("WARNING the file " + str(filename) +
                  ".csv already exists and is open.")
            print("Was unable to save the file")

    else:
        try:
            with open(filename, 'w', encoding='UTF8', newline="") as f:
                writer = csv.writer(f)

                for name, value in zip(names, values):
                    data = [name]+value
                    writer.writerow(data)
        except PermissionError:

            print("WARNING the file " + str(filename)
                  + ".csv already exists and is open.")
            print("Was unable to save the file")



# Copyright (C) 2020 - 2022
# Bardhyl Miftari, Mathias Berger, Hatim Djelassi, Damien Ernst,
# University of Liege .
# Licensed under the MIT License (see LICENSE file).

from .classes import Program
from .utils import error_, get_only_objects_in_nested_dict_layer

import numpy as np  # type: ignore
from scipy.sparse import coo_matrix  # type: ignore
import multiprocessing as mp


def get_flat_nodes_edges_ordered(nodes, hyperedges, without_hyperedges=False):
    all_nodes_edges = []
    for node in nodes:
        all_nodes_edges.append(node)
        sub_nodes = node.get_sub_nodes()
        sub_edges = node.get_sub_hyperedges()
        all_sub_nodes_edges = get_flat_nodes_edges_ordered(sub_nodes,
                                                           sub_edges,
                                                           without_hyperedges)
        all_nodes_edges += all_sub_nodes_edges

    if not without_hyperedges:
        all_nodes_edges += hyperedges

    return all_nodes_edges


def get_flat_nodes_edges_and_definition(nodes, hyperedges,
                                        definitions, accumulator_dict):
    node_definition_tuples = []
    edge_definition_tuples = []
    for node in nodes:
        node_name = node.get_name()
        sub_nodes = node.get_sub_nodes()
        sub_edges = node.get_sub_hyperedges()
        accumulator_dict[node_name] = \
            get_only_objects_in_nested_dict_layer(definitions[node_name])
        node_definition_tuples.append([node, accumulator_dict.copy()])
        sub_node_tuples, sub_edge_tuples = \
            get_flat_nodes_edges_and_definition(sub_nodes, sub_edges,
                                                definitions[node_name],
                                                accumulator_dict)
        node_definition_tuples += sub_node_tuples
        edge_definition_tuples += sub_edge_tuples
        accumulator_dict.pop(node_name)

    for edge in hyperedges:
        edge_name = edge.get_name()
        accumulator_dict[edge_name] = \
            get_only_objects_in_nested_dict_layer(definitions[edge_name])
        edge_definition_tuples.append([edge, accumulator_dict.copy()])
        accumulator_dict.pop(edge_name)

    return node_definition_tuples, edge_definition_tuples


def extend_node_factors_mapping(list_elements_definitions_tuple):
    list_elements, definitions = list_elements_definitions_tuple
    all_triplet_matrix_independent_terms_type = []
    for element in list_elements:
        element.extend(definitions)
        sparse, independent_terms, extension_type = \
            element.sparse, element.independent_terms, element.extension_type

        all_triplet_matrix_independent_terms_type.append([sparse,
                                                          independent_terms,
                                                          extension_type])
    return all_triplet_matrix_independent_terms_type


def extend_factor_on_multiple_processes(root: Program,
                                        definitions,
                                        nb_processes) -> None:
    nodes = root.get_nodes()
    hyperlinks = root.get_links()
    all_tuples = []
    accumulator_dict = {"global": definitions["global"], "T": definitions["T"]}
    node_definition_tuples, edge_definition_tuples = \
        get_flat_nodes_edges_and_definition(nodes, hyperlinks, definitions,
                                            accumulator_dict)
    for node, node_definition in node_definition_tuples:
        obj_fact_list = node.get_objective_factors()
        constr_fact_list = node.get_constraint_factors()
        all_tuples.append([obj_fact_list + constr_fact_list, node_definition])

    for link, link_definition in edge_definition_tuples:
        constr_fact_list = link.get_constraint_factors()
        all_tuples.append([constr_fact_list, link_definition])

    with mp.Pool(processes=nb_processes) as pool:
        results = pool.map(extend_node_factors_mapping, all_tuples)
    # pool.close()

    for i, node_factors in enumerate(all_tuples):
        node_results = results[i]
        factor_list, node_dict = node_factors
        for j, factor in enumerate(factor_list):
            sparse, independent_terms, extension_type = node_results[j]
            factor.sparse = sparse
            factor.extension_type = extension_type
            factor.independent_terms = independent_terms


def extend_factor(root: Program, definitions) -> None:
    nodes = root.get_nodes()
    hyperlinks = root.get_links()
    accumulator_dict = {"global": definitions["global"], "T": definitions["T"]}
    node_definition_tuples, edge_definition_tuples = \
        get_flat_nodes_edges_and_definition(nodes, hyperlinks, definitions,
                                            accumulator_dict)
    for node, node_definition in node_definition_tuples:
        obj_fact_list = node.get_objective_factors()
        for i, object_fact in enumerate(obj_fact_list):
            object_fact.extend(node_definition)
        constr_fact_list = node.get_constraint_factors()
        for constr_fact in constr_fact_list:
            constr_fact.extend(node_definition)

    for link, link_definition in edge_definition_tuples:
        constr_fact_list = link.get_constraint_factors()
        for constr_fact in constr_fact_list:
            constr_fact.extend(link_definition)


def matrix_generation_c(root: Program) -> tuple:
    """
    matrix_generationC function: takes as input a program object and returns
    a coo_matrix of the different objectives flatten. In other words, returns
    the different objectives as a matrix : min C*x where each line of C
    corresponds to one objective.
    INPUT:  Program object
    OUTPUT: C -> Sparse coo matrix of the objective function
            objective_map -> Mapping to check which objective relates
                             to which node
    """

    nb_variables = root.get_nb_var_index()
    nodes = root.get_nodes()
    nodes = get_flat_nodes_edges_ordered(nodes, [], without_hyperedges=True)
    length_values = 0
    length_independent_terms = 0
    number_of_objectives = 0
    for node in nodes:
        objective_factors_list = node.get_objective_factors()
        for objective_index, objective_factor in \
                enumerate(objective_factors_list):
            internal_sparse: coo_matrix = objective_factor.sparse
            if internal_sparse is None:
                continue
            number_objective, _ = internal_sparse.shape
            independent_terms = objective_factor.independent_terms
            values = internal_sparse.data

            length_values += len(values)
            number_of_objectives += number_objective
            length_independent_terms += len(independent_terms)

    all_values = np.zeros(length_values)
    all_rows = np.zeros(length_values)
    all_columns = np.zeros(length_values)
    indep_terms = np.zeros(length_independent_terms)
    values_offset = 0
    independent_terms_offset = 0
    objective_offset = 0

    for node in nodes:
        node_objectives = {}
        number_node_objectives = 0
        objective_factors_list = node.get_objective_factors()

        for objective_index, objective_factor in \
                enumerate(objective_factors_list):

            internal_sparse: coo_matrix = objective_factor.sparse
            if internal_sparse is None:
                continue
            number_objectives, _ = internal_sparse.shape
            optimization_type = objective_factor.extension_type
            values, rows, columns = \
                internal_sparse.data, internal_sparse.row, internal_sparse.col
            independent_terms = objective_factor.independent_terms
            name_objective = objective_factor.get_name()
            rows += objective_offset

            number_of_values = len(values)
            number_of_independent_terms = len(independent_terms)

            obj_data = dict()
            obj_data["type"] = optimization_type
            obj_data["indexes"] = np.arange(objective_offset,
                                            objective_offset
                                            + number_objectives)
            if name_objective is not None:
                obj_data["name"] = name_objective
            node_objectives[objective_index] = obj_data

            if optimization_type == "max":
                values = - values
                independent_terms = - independent_terms

            all_values[values_offset:values_offset + number_of_values] = values
            all_columns[values_offset:values_offset + number_of_values] = \
                columns
            all_rows[values_offset:values_offset + number_of_values] = rows
            indep_terms[independent_terms_offset:
                        independent_terms_offset
                        + number_of_independent_terms] = independent_terms
            values_offset += number_of_values
            independent_terms_offset += number_of_independent_terms
            objective_offset += number_objectives
            number_node_objectives += number_objectives

        node.set_objectives_data(node_objectives)
        node.nb_objective_matrix = number_node_objectives
        node.objective_list = None

    if len(all_rows) == 0:
        error_("ERROR: no valid objective defined")

    sparse_matrix = coo_matrix((all_values, (all_rows, all_columns)),
                               shape=(number_of_objectives, nb_variables))
    return sparse_matrix, indep_terms


def matrix_generation_a_b(root: Program) -> tuple:
    """
    matrix_generationAb function: takes as input a program object and
    returns a tuple composed of a sparse matrix of constraints A and
    a vector of independent terms b.
    INPUT:  Program object
    OUTPUT: A -> Sparse coo matrix of the constraints
            b -> Np.ndarray of the independent term of each constraint
    """

    number_of_variables = root.get_nb_var_index()
    nodes = root.get_nodes()
    hyperlinks = root.get_links()
    graph_elements = get_flat_nodes_edges_ordered(nodes, hyperlinks,
                                                  without_hyperedges=False)
    length_values = 0
    length_independent_terms = 0
    number_of_constraints = 0

    for obj in graph_elements:
        constr_fact_list = obj.get_constraint_factors()
        for constr_fact in constr_fact_list:
            internal_sparse: coo_matrix = constr_fact.sparse
            if internal_sparse is None:
                continue
            number_constraints, _ = internal_sparse.shape
            independent_terms = constr_fact.independent_terms
            values = internal_sparse.data

            length_values += len(values)
            number_of_constraints += number_constraints
            length_independent_terms += len(independent_terms)

    all_values = np.zeros(length_values)
    all_rows = np.zeros(length_values)
    all_columns = np.zeros(length_values)
    all_rhs = np.zeros(length_independent_terms)
    values_offset = 0
    independent_terms_offset = 0
    constraint_offset = 0

    for obj in graph_elements:
        constr_fact_list = obj.get_constraint_factors()
        number_node_constraints = 0
        factor_mapping = {}
        for constr_fact in constr_fact_list:

            internal_sparse: coo_matrix = constr_fact.sparse
            if internal_sparse is None:
                continue
            number_constraints, _ = internal_sparse.shape
            sign = constr_fact.extension_type
            values, rows, columns = \
                internal_sparse.data, internal_sparse.row, internal_sparse.col
            independent_terms = constr_fact.independent_terms
            rows += constraint_offset
            number_of_values = len(values)
            number_of_independent_terms = len(independent_terms)

            if sign == ">=":
                # Do -c<=-b
                values = -values
                independent_terms = - independent_terms

            all_values[values_offset:values_offset + number_of_values] = values
            all_columns[values_offset:values_offset + number_of_values] = \
                columns
            all_rows[values_offset:values_offset + number_of_values] = rows
            all_rhs[independent_terms_offset:
                    independent_terms_offset
                    + number_of_independent_terms] = independent_terms

            values_offset += number_of_values
            independent_terms_offset += number_of_independent_terms
            constraint_offset += number_constraints
            number_node_constraints += number_constraints

            if constr_fact.get_name():
                factor_mapping[constr_fact.get_name()] \
                    = slice(constraint_offset - number_constraints,
                            constraint_offset)
        if factor_mapping:
            obj.set_constraints_data(factor_mapping)
        obj.free_factors_constraints()
        obj.c_triplet_list = None
        obj.nb_constraint_matrix = number_node_constraints
    root.link_constraints = None
    sparse_matrix = coo_matrix((all_values, (all_rows, all_columns)),
                               shape=(number_of_constraints,
                                      number_of_variables))

    return sparse_matrix, all_rhs
